@@ -1,11 +1,11 @@
 # postgresql-lib.pl
 # Common PostgreSQL functions
 # XXX updating date field
-# XXX access control and schema tables
 
-do '../web-lib.pl';
+BEGIN { push(@INC, ".."); };
+use WebminCore;
 &init_config();
-require '../ui-lib.pl';
+do 'view-lib.pl';
 if ($config{'plib'}) {
 	$ENV{$gconfig{'ld_env'}} .= ':' if ($ENV{$gconfig{'ld_env'}});
 	$ENV{$gconfig{'ld_env'}} .= $config{'plib'};
@@ -14,35 +14,45 @@ if ($config{'psql'} =~ /^(.*)\/bin\/psql$/ && $1 ne '' && $1 ne '/usr') {
 	$ENV{$gconfig{'ld_env'}} .= ':' if ($ENV{$gconfig{'ld_env'}});
 	$ENV{$gconfig{'ld_env'}} .= "$1/lib";
 	}
+$pg_shadow_cols = "usename,usesysid,usecreatedb,usesuper,usecatupd,passwd,valuntil";
 
 if ($module_info{'usermin'}) {
 	# Login and password is set by user in Usermin, and the module always
 	# runs as the Usermin user
 	&switch_to_remote_user();
 	&create_user_config_dirs();
-	$postgres_login = $userconfig{'login'};
-	$postgres_pass = $userconfig{'pass'};
-	$postgres_sameunix = 0;
+	&set_login_pass(0, $userconfig{'login'}, $userconfig{'pass'});
 	%access = ( 'backup' => 1,
 		    'restore' => 1,
 		    'tables' => 1,
 		    'cmds' => 1, );
 	$max_dbs = $userconfig{'max_dbs'};
+	$commands_file = "$user_module_config_directory/commands";
+	%displayconfig = %userconfig;
 	}
 else {
 	# Login and password is determined by ACL in Webmin
 	%access = &get_module_acl();
 	if ($access{'user'} && !$use_global_login) {
-		$postgres_login = $access{'user'};
-		$postgres_pass = $access{'pass'};
-		$postgres_sameunix = $access{'sameunix'};
+		&set_login_pass(
+			$access{'sameunix'}, $access{'user'}, $access{'pass'});
 		}
 	else {
-		$postgres_login = $config{'login'};
-		$postgres_pass = $config{'pass'};
-		$postgres_sameunix = $config{'sameunix'};
+		&set_login_pass(
+			$config{'sameunix'}, $config{'login'}, $config{'pass'});
 		}
 	$max_dbs = $config{'max_dbs'};
+	$commands_file = "$module_config_directory/commands";
+	%displayconfig = %config;
+	}
+foreach my $hba (split(/\t+/, $config{'hba_conf'})) {
+	if ($hba =~ /\*|\?/) {
+		($hba) = glob($hba);
+		}
+	if ($hba && -r $hba) {
+		$hba_conf_file = $hba;
+		last;
+		}
 	}
 $cron_cmd = "$module_config_directory/backup.pl";
 
@@ -61,22 +71,15 @@ EOF
 sub is_postgresql_running
 {
 local $temp = &transname();
-local $host = $config{'host'} ? "-h $config{'host'}" : "";
-$host .= " -p $config{'port'}" if ($config{'port'});
-local $cmd;
-local $out;
-if ($postgres_login) {
-	open(TEMP, ">$temp");
-	print TEMP "$postgres_login\n$postgres_pass\n";
-	close(TEMP);
-	$cmd = "$config{'psql'} -u -c '' $host $config{'basedb'} <$temp";
-	}
-else {
-	$cmd = "$config{'psql'} -c '' $host $config{'basedb'}";
-	}
+local $cmd = &quote_path($config{'psql'}).
+	     &host_port_flags().
+	     (!&supports_pgpass() ? " -u" : " -U $postgres_login").
+	     " -c ''".
+	     " ".$config{'basedb'};
 if ($postgres_sameunix && defined(getpwnam($postgres_login))) {
 	$cmd = "su $postgres_login -c ".quotemeta($cmd);
 	}
+$cmd = &command_with_login($cmd);
 if (&foreign_check("proc")) {
 	&foreign_require("proc", "proc-lib.pl");
 	if (defined(&proc::close_controlling_pty)) {
@@ -111,12 +114,12 @@ return wantarray ? ($rv, $out) : $rv;
 # get_postgresql_version([from-command])
 sub get_postgresql_version
 {
-local ($fromcmd) = @_;
-local $main::error_must_die = 1;
+my ($fromcmd) = @_;
 return $postgresql_version_cache if (defined($postgresql_version_cache));
-local $rv;
+my $rv;
 if (!$fromcmd) {
 	eval {
+		local $main::error_must_die = 1;
 		local $v = &execute_sql_safe($config{'basedb'},
 					     'select version()');
 		$v = $v->{'data'}->[0]->[0];
@@ -126,7 +129,8 @@ if (!$fromcmd) {
 		};
 	}
 if (!$rv || $@) {
-	local $out = `$config{'psql'} -V 2>&1 </dev/null`;
+	my $out = &backquote_command(
+		&quote_path($config{'psql'})." -V 2>&1 <$null_file");
 	$rv = $out =~ /\s([0-9\.]+)/ ? $1 : undef;
 	}
 $postgresql_version_cache = $rv;
@@ -143,8 +147,8 @@ return &get_postgresql_version() >= 7.3;
 sub list_databases
 {
 local $force_nodbi = 1;
-local $t = &execute_sql_safe($config{'basedb'}, 'select * from pg_database order by datname');
-return map { $_->[0] } @{$t->{'data'}};
+local $t = &execute_sql_safe($config{'basedb'}, 'select datname from pg_database order by datname');
+return sort { lc($a) cmp lc($b) } map { $_->[0] } @{$t->{'data'}};
 }
 
 # supports_schemas(database)
@@ -160,12 +164,13 @@ return $t->{'data'}->[0]->[0] ? 1 : 0;
 sub list_tables
 {
 if (&supports_schemas($_[0])) {
-	local $t = &execute_sql_safe($_[0], 'select schemaname,tablename from pg_tables where tablename not like \'pg_%\' and tablename not like \'sql_%\' order by tablename');
-	return map { ($_->[0] eq "public" ? "" : $_->[0].".").$_->[1] } @{$t->{'data'}};
+	local $t = &execute_sql_safe($_[0], 'select schemaname,tablename from pg_tables order by tablename');
+	return map { ($_->[0] eq "public" ? "" : $_->[0].".").$_->[1] }
+		   grep { $_->[1] !~ /^(pg|sql)_/ } @{$t->{'data'}};
 	}
 else {
-	local $t = &execute_sql_safe($_[0], 'select tablename from pg_tables where tablename not like \'pg_%\' and tablename not like \'sql_%\' order by tablename');
-	return map { $_->[0] } @{$t->{'data'}};
+	local $t = &execute_sql_safe($_[0], 'select tablename from pg_tables order by tablename');
+	return map { $_->[0] } grep { $_->[0] !~ /^(pg|sql)_/ } @{$t->{'data'}};
 	}
 }
 
@@ -213,7 +218,8 @@ if (&supports_schemas($_[0])) {
 
 	# Work out which fields are the primary key
 	if (&supports_indexes()) {
-		local ($keyidx) = grep { $_ eq $_[1]."_pkey" }
+		local ($keyidx) = grep { $_ eq $_[1]."_pkey" ||
+					 $_ eq "pk_".$_[1] }
 				       &list_indexes($_[0]);
 		if ($keyidx) {
 			local $istr = &index_structure($_[0], $keyidx);
@@ -263,6 +269,17 @@ sub execute_sql_safe
 {
 local $sql = $_[1];
 local @params = @_[2..$#_];
+if ($gconfig{'debug_what_sql'}) {
+	# Write to Webmin debug log
+	local $params;
+	for(my $i=0; $i<@params; $i++) {
+		$params .= " ".$i."=".$params[$i];
+		}
+	&webmin_debug_log('SQL', "db=$_[0] sql=$sql".$params);
+	}
+if ($sql !~ /^\s*\\/ && !$main::disable_postgresql_escaping) {
+	$sql =~ s/\\/\\\\/g;
+	}
 if ($driver_handle &&
     $sql !~ /^\s*(create|drop)\s+database/ && $sql !~ /^\s*\\/ &&
     !$force_nodbi) {
@@ -271,14 +288,16 @@ if ($driver_handle &&
 	local $cstr = "dbname=$_[0]";
 	$cstr .= ";host=$config{'host'}" if ($config{'host'});
 	$cstr .= ";port=$config{'port'}" if ($config{'port'});
+	local $sslmode = $config{'sslmode'};
+	$sslmode =~ s/_/-/g;
+	$cstr .= ";sslmode=$sslmode" if ($sslmode);
 	local @uinfo;
 	if ($postgres_sameunix &&
-	    defined(@uinfo = getpwnam($postgres_login))) {
+	    (@uinfo = getpwnam($postgres_login))) {
 		# DBI call which must run in subprocess
 		pipe(OUTr, OUTw);
 		if (!($pid = fork())) {
-			($(, $)) = ( $uinfo[3], $uinfo[3] );
-			($>, $<) = ( $uinfo[2], $uinfo[2] );
+			&switch_to_unix_user(\@uinfo);
 			close(OUTr);
 			local $dbh = $driver_handle->connect($cstr,
 					$postgres_login, $postgres_pass);
@@ -348,13 +367,16 @@ if ($driver_handle &&
 else {
 	# Check for a \ command
         my $break_f = 0 ;
-        if ( $sql =~ /^\s*\\/ ) {
-            $break_f = 1 ;
-            if ( $sql !~ /^\s*\\copy\s+/ &&
-                 $sql !~ /^\s*\\i\s+/ ) {
-                &error ( &text ( 'r_command', ) ) ;
-            }
-        }
+	if ($sql =~ /^\s*\\l\s*$/) {
+		# \l command to list encodings needs no special handling
+		}
+        elsif ($sql =~ /^\s*\\/ ) {
+		$break_f = 1 ;
+		if ($sql !~ /^\s*\\copy\s+/ &&
+                    $sql !~ /^\s*\\i\s+/) {
+			&error ( &text ( 'r_command', ) ) ;
+			}
+		}
 
 	if (@params) {
 		# Sub in ? parameters
@@ -378,19 +400,19 @@ else {
 		}
 
 	# Call the psql program
-	local $host = $config{'host'} ? "-h $config{'host'}" : "";
-	$host .= " -p $config{'port'}" if ($config{'port'});
-	local $cmd = $config{'psql'}.
+	local $cmd = &quote_path($config{'psql'})." --html".
+		     &host_port_flags().
 		     (!&supports_pgpass() ? " -u" : " -U $postgres_login").
-		     " -c ".quotemeta($sql)." $host $_[0]";
+		     " -c ".&quote_path($sql).
+		     " ".$_[0];
 	if ($postgres_sameunix && defined(getpwnam($postgres_login))) {
-		$cmd = "su $postgres_login -c ".quotemeta($cmd);
+		$cmd = &command_as_user($postgres_login, 0, $cmd);
 		}
 	$cmd = &command_with_login($cmd);
 
 	delete($ENV{'LANG'});		# to force output to english
 	delete($ENV{'LANGUAGE'});
-        if ( $break_f == 0 ) {
+        if ($break_f == 0) {
 		# Running a normal SQL command, not one with a \
 		#$ENV{'PAGER'} = "cat";
 		if (&foreign_check("proc")) {
@@ -405,9 +427,9 @@ else {
 		local ($line, $rv, @data);
 		do {
 			$line = <OUT>;
-			last if (!defined($line));
 			} while($line =~ /^(username|password|user name):/i ||
-				$line =~ /(warning|notice):/i || $line !~ /\S/);
+				$line =~ /(warning|notice):/i ||
+			        $line !~ /\S/ && defined($line));
 		unlink($temp);
 		if ($line =~ /^ERROR:\s+(.*)/ || $line =~ /FATAL.*:\s+(.*)/) {
 			&error(&text('esql', "<tt>$sql</tt>", "<tt>$1</tt>"));
@@ -418,44 +440,28 @@ else {
 				     "<tt>$config{'psql'} failed</tt>"));
 			}
 		else {
-			local $dash = <OUT>;
-			if ($dash =~ /^\s*\+\-/) {
-				# mysql-style output
-				$line = <OUT>;
-				$line =~ s/^[\s\|]+//; $line =~ s/[\s\|]+$//;
-				local @titles = split(/\|/, $line);
-				map { s/^\s+//; s/\s+$// } @titles;
-				$line = <OUT>;	# skip useless dashes
-				while(1) {
-					$line = <OUT>;
-					last if (!$line || $line =~ /^\s*\+/);
-					$line =~ s/^[\s\|]+//;
-					$line =~ s/[\s\|]+$//;
-					local @row = split(/\|/, $line);
-					map { s/^\s+//; s/\s+$// } @row;
-					push(@data, \@row);
+			# Read HTML-format output
+			local $row;
+			local @data;
+			while($line = <OUT>) {
+				if ($line =~ /^\s*<tr>/) {
+					# Start of a row
+					$row = [ ];
 					}
-				$rv = { 'titles' => \@titles, 'data' => \@data };
-				}
-			elsif ($dash !~ /^-/) {
-				# no output, such as from an insert
-				$rv = undef;
-				}
-			else {
-				# psql-style output
-				local @titles = split(/\|/, $line);
-				map { s/^\s+//; s/\s+$// } @titles;
-				while(1) {
-					$line = <OUT>;
-					last if (!$line ||
-						 $line =~ /^\(\d+\s+\S+\)/);
-					local @row = split(/\|/, $line);
-					map { s/^\s+//; s/\s+$// } @row;
-					push(@data, \@row);
+				elsif ($line =~ /^\s*<\/tr>/) {
+					# End of a row
+					push(@data, $row);
+					$row = undef;
+					}
+				elsif ($line =~ /^\s*<(td|th)[^>]*>(.*)<\/(td|th)>/) {
+					# Value in a row
+					local $v = $2;
+					$v =~ s/<br>/\n/g;
+					push(@$row, &entities_to_ascii($v));
 					}
-				$rv = { 'titles' => \@titles,
-					'data' => \@data };
 				}
+			$rv = { 'titles' => shift(@data),
+				'data' => \@data };
 			}
 		close(OUT);
 		return $rv;
@@ -466,10 +472,10 @@ else {
 
 		$emsgf = &transname();
 		$rc = &system_logged ( "$cmd >$emsgf 2>&1");
-		$emsg  = `cat $emsgf` ;
-		unlink ( $emsgf ) ;
+		$emsg  = &read_file_contents($emsgf);
+		&unlink_file($emsgf) ;
 		if ($rc) {
-			&error ( "<pre>$emsg</pre>" );
+			&error("<pre>$emsg</pre>");
 			}
 		else {
 			@titles = ( "     Command Invocation      " ) ;
@@ -539,7 +545,7 @@ else {
 sub get_hba_config
 {
 local $lnum = 0;
-open(HBA, $config{'hba_conf'});
+open(HBA, "<".$hba_conf_file);
 while(<HBA>) {
 	s/\r|\n//g;
 	s/^\s*#.*$//g;
@@ -609,7 +615,7 @@ return @rv;
 # create_hba(&hba, version)
 sub create_hba
 {
-local $lref = &read_file_lines($config{'hba_conf'});
+local $lref = &read_file_lines($hba_conf_file);
 push(@$lref, &hba_line($_[0], $_[1]));
 &flush_file_lines();
 }
@@ -617,7 +623,7 @@ push(@$lref, &hba_line($_[0], $_[1]));
 # delete_hba(&hba, version)
 sub delete_hba
 {
-local $lref = &read_file_lines($config{'hba_conf'});
+local $lref = &read_file_lines($hba_conf_file);
 splice(@$lref, $_[0]->{'line'}, 1);
 &flush_file_lines();
 }
@@ -625,7 +631,7 @@ splice(@$lref, $_[0]->{'line'}, 1);
 # modify_hba(&hba, version)
 sub modify_hba
 {
-local $lref = &read_file_lines($config{'hba_conf'});
+local $lref = &read_file_lines($hba_conf_file);
 splice(@$lref, $_[0]->{'line'}, 1, &hba_line($_[0], $_[1]));
 &flush_file_lines();
 }
@@ -633,7 +639,7 @@ splice(@$lref, $_[0]->{'line'}, 1, &hba_line($_[0], $_[1]));
 # swap_hba(&hba1, &hba2)
 sub swap_hba
 {
-local $lref = &read_file_lines($config{'hba_conf'});
+local $lref = &read_file_lines($hba_conf_file);
 local $line0 = $lref->[$_[0]->{'line'}];
 local $line1 = $lref->[$_[1]->{'line'}];
 $lref->[$_[1]->{'line'}] = $line0;
@@ -692,10 +698,11 @@ return $_[0]->{'type'} eq 'text' || $_[0]->{'type'} eq 'bytea';
 # HUP postmaster if running, so that hosts file changes take effect
 sub restart_postgresql
 {
-if (open(PID, $config{'pid_file'})) {
-	($pid = <PID>) =~ s/\r|\n//g;
-	close(PID);
-	&kill_logged('HUP', $pid) if ($pid);
+foreach my $pidfile (glob($config{'pid_file'})) {
+	local $pid = &check_pid_file($pidfile);
+	if ($pid) {
+		&kill_logged('HUP', $pid);
+		}
 	}
 }
 
@@ -703,15 +710,23 @@ if (open(PID, $config{'pid_file'})) {
 # Does strftime-style date substitutions on a filename, if enabled
 sub date_subs
 {
+local ($path) = @_;
+local $rv;
 if ($config{'date_subs'}) {
         eval "use POSIX";
 	eval "use posix" if ($@);
         local @tm = localtime(time());
-        return strftime($_[0], @tm);
+        &clear_time_locale();
+        $rv = strftime($path, @tm);
+        &reset_time_locale();
         }
 else {
-        return $_[0];
+	$rv = $path;
         }
+if ($config{'webmin_subs'}) {
+	$rv = &substitute_template($rv, { });
+	}
+return $rv;
 }
 
 # execute_before(db, handle, escape, path, db-for-config)
@@ -748,6 +763,19 @@ if ($cmd) {
 return 1;
 }
 
+# make_backup_dir(directory)
+# Create a directory that PostgreSQL can backup into
+sub make_backup_dir
+{
+local ($dir) = @_;
+if (!-d $dir) {
+	&make_dir($dir, 0755);
+	if ($postgres_sameunix && defined(getpwnam($postgres_login))) {
+		&set_ownership_permissions($postgres_login, undef, undef, $dir);
+		}
+	}
+}
+
 sub quote_table
 {
 local @tn = split(/\./, $_[0]);
@@ -759,23 +787,28 @@ sub quotestr
 return "\"$_[0]\"";
 }
 
-# execute_sql_file(database, file, [user, pass])
+# execute_sql_file(database, file, [user, pass], [unix-user])
 # Executes some file of SQL statements, and returns the exit status and output
 sub execute_sql_file
 {
+local ($db, $file, $user, $pass, $unixuser) = @_;
 if (&is_readonly_mode()) {
 	return (0, undef);
 	}
-local ($db, $file, $user, $pass) = @_;
 if (!defined($user)) {
 	$user = $postgres_login;
 	$pass = $postgres_pass;
 	}
-local $cmd = "$config{'psql'} -f $file".
+local $cmd = &quote_path($config{'psql'})." -f ".&quote_path($file).
+	     &host_port_flags().
 	     (&supports_pgpass() ? " -U $user" : " -u").
-	     ($config{'host'} ? " -h $config{'host'}" : "").
-	     ($config{'port'} ? " -h $config{'port'}" : "").
-	     " $db";
+	     " ".$db;
+if ($postgres_sameunix && defined(getpwnam($postgres_login))) {
+	$cmd = &command_as_user($postgres_login, 0, $cmd);
+	}
+elsif ($unixuser && $unixuser ne 'root' && $< == 0) {
+	$cmd = &command_as_user($unixuser, 0, $cmd);
+	}
 $cmd = &command_with_login($cmd, $user, $pass);
 local $out = &backquote_logged("$cmd 2>&1");
 return ($out =~ /ERROR/i ? 1 : 0, $out);
@@ -785,14 +818,14 @@ return ($out =~ /ERROR/i ? 1 : 0, $out);
 # Outputs a table that is split into two parts
 sub split_table
 {
-local $mid = int((@{$_[1]}+1) / 2);
+local $mid = int((@{$_[2]}+1) / 2);
 local ($i, $j);
 print "<table width=100%><tr>\n";
-foreach $s ([0, $mid-1], [$mid, @{$_[1]}-1]) {
+foreach $s ([0, $mid-1], [$mid, @{$_[2]}-1]) {
 	print "<td width=50% valign=top>\n";
 
 	# Header
-	local @tds = ( "width=5" );
+	local @tds = $_[1] ? ( "width=5" ) : ( );
 	if ($s->[0] <= $s->[1]) {
 		local @hcols;
 		foreach $t (@{$_[0]}) {
@@ -807,7 +840,12 @@ foreach $s ([0, $mid-1], [$mid, @{$_[1]}-1]) {
 		for($j=4; $j<@_; $j++) {
 			push(@cols, $_[$j]->[$i]);
 			}
-		print &ui_checked_columns_row(\@cols, \@tds, "d", $_[1]->[$i]);
+		if ($_[1]) {
+			print &ui_checked_columns_row(\@cols, \@tds, "d", $_[1]->[$i]);
+			}
+		else {
+			print &ui_columns_row(\@cols, \@tds);
+			}
 		}
 	if ($s->[0] <= $s->[1]) {
 		print &ui_columns_end();
@@ -843,6 +881,15 @@ return 1;
 # or undef on success.
 sub start_postgresql
 {
+if ($gconfig{'os_type'} eq 'windows' && &foreign_check("init")) {
+	# On Windows, always try to sc start the pgsql- service
+	&foreign_require("init", "init-lib.pl");
+	local ($pg) = grep { $_->{'name'} =~ /^pgsql-/ }
+			   &init::list_win32_services();
+	if ($pg) {
+		return &init::start_win32_service($pg->{'name'});
+		}
+	}
 local $temp = &transname();
 local $rv = &system_logged("($config{'start_cmd'}) >$temp 2>&1");
 local $out = `cat $temp`; unlink($temp);
@@ -858,6 +905,15 @@ return undef;
 # or undef on success.
 sub stop_postgresql
 {
+if ($gconfig{'os_type'} eq 'windows' && &foreign_check("init")) {
+	# On Windows, always try to sc stop the pgsql- service
+	&foreign_require("init", "init-lib.pl");
+	local ($pg) = grep { $_->{'name'} =~ /^pgsql-/ }
+			   &init::list_win32_services();
+	if ($pg) {
+		return &init::stop_win32_service($pg->{'name'});
+		}
+	}
 if ($config{'stop_cmd'}) {
 	local $out = &backquote_logged("$config{'stop_cmd'} 2>&1");
 	if ($? || $out =~ /failed|error/i) {
@@ -865,13 +921,18 @@ if ($config{'stop_cmd'}) {
 		}
 	}
 else {
-	local $pid;
-	open(PID, $config{'pid_file'});
-	($pid = <PID>) =~ s/\r|\n//g;
-	close(PID);
-	$pid || return &text('stop_epidfile', "<tt>$config{'pid_file'}</tt>");
-	&kill_logged('TERM', $pid) ||
-		return &text('stop_ekill', "<tt>$pid</tt>", "<tt>$!</tt>");
+	local $pidcount = 0;
+	foreach my $pidfile (glob($config{'pid_file'})) {
+		local $pid = &check_pid_file($pidfile);
+		if ($pid) {
+			&kill_logged('TERM', $pid) ||
+				return &text('stop_ekill', "<tt>$pid</tt>",
+					     "<tt>$!</tt>");
+			$pidcount++;
+			}
+		}
+	$pidcount || return &text('stop_epidfile',
+				  "<tt>$config{'pid_file'}</tt>");
 	}
 return undef;
 }
@@ -951,6 +1012,9 @@ if ($info->{'create'} =~ /using\s+(\S+)\s/) {
 	}
 if ($info->{'create'} =~ /\((.*)\)/) {
 	$info->{'cols'} = [ split(/\s*,\s*/, $1) ];
+	foreach my $c (@{$info->{'cols'}}) {
+		$c =~ s/^"(.*)"$/$1/;
+		}
 	}
 
 return $info;
@@ -1060,59 +1124,84 @@ return $config{'host'} eq '' || $config{'host'} eq 'localhost' ||
        &to_ipaddress($config{'host'}) eq &to_ipaddress(&get_system_hostname());
 }
 
-# backup_database(database, dest-path, format, [&only-tables])
+# backup_database(database, dest-path, format, [&only-tables], [run-as-user],
+# 		  [compress-mode])
 # Executes the pg_dump command to backup the specified database to the
 # given destination path. Returns undef on success, or an error message
 # on failure.
 sub backup_database
 {
-local ($db, $path, $format, $tables) = @_;
-local $tablesarg = join(" ", map { " -t ".quotemeta($_) } @$tables);
-local $cmd = $config{'dump_cmd'}.
+my ($db, $path, $format, $tables, $user, $compress) = @_;
+my $tablesarg = join(" ", map { " -t ".quotemeta('"'.$_.'"') } @$tables);
+my $writer;
+if ($compress == 0) {
+        $writer = "cat >".quotemeta($path);
+        }
+elsif ($compress == 1) {
+        $writer = "gzip -c >".quotemeta($path);
+        }
+elsif ($compress == 2) {
+        $writer = "bzip2 -c >".quotemeta($path);
+        }
+my $cmd = &quote_path($config{'dump_cmd'}).
+	     &host_port_flags().
 	     (!$postgres_login ? "" :
 	      &supports_pgpass() ? " -U $postgres_login" : " -u").
-	     ($config{'host'} ? " -h $config{'host'}" : "").
 	     ($format eq 'p' ? "" : " -b").
 	     $tablesarg.
-	     " -F$format -f ".quotemeta($path)." $db";
+	     " -F$format $db | $writer";
 if ($postgres_sameunix && defined(getpwnam($postgres_login))) {
+	# Postgres connections have to be made as the 'postgres' Unix user
 	$cmd = &command_as_user($postgres_login, 0, $cmd);
 	}
+elsif ($user) {
+	# Run as a specific Unix user
+	$cmd = &command_as_user($user, 0, $cmd);
+	}
 $cmd = &command_with_login($cmd);
-local $out = &backquote_logged("$cmd 2>&1");
+my $out = &backquote_logged("$cmd 2>&1");
 if ($? || $out =~ /could not|error|failed/i) {
 	return $out;
 	}
 return undef;
 }
 
-# restore_database(database, source-path, only-data, clear-db)
+# restore_database(database, source-path, only-data, clear-db, [&only-tables],
+# 		   [login, password])
 # Restores the contents of a PostgreSQL backup into the specified database.
 # Returns undef on success, or an error message on failure.
 sub restore_database
 {
-local ($db, $path, $only, $clean) = @_;
-local $cmd = $config{'rstr_cmd'}.
-	     (!$postgres_login ? "" :
-	      &supports_pgpass() ? " -U $postgres_login" : " -u").
-	     ($config{'host'} ? " -h $config{'host'}" : "").
+my ($db, $path, $only, $clean, $tables, $login, $pass) = @_;
+my $tablesarg = join(" ", map { " -t ".quotemeta('"'.$_.'"') } @$tables);
+$login ||= $postgres_login;
+$pass ||= $postgres_pass;
+my $cmd = &quote_path($config{'rstr_cmd'}).
+	     &host_port_flags().
+	     (!$login ? "" :
+	      &supports_pgpass() ? " -U $login" : " -u").
 	     ($only ? " -a" : "").
 	     ($clean ? " -c" : "").
-	     " -d $in{'db'} ".quotemeta($path);
-if ($postgres_sameunix && defined(getpwnam($postgres_login))) {
-	$cmd = &command_as_user($postgres_login, 0, $cmd);
+	     $tablesarg.
+	     " -d $db ".&quote_path($path);
+if ($postgres_sameunix && defined(getpwnam($login))) {
+	$cmd = &command_as_user($login, 0, $cmd);
 	}
-$cmd = &command_with_login($cmd);
-local $out = &backquote_logged("$cmd 2>&1");
+$cmd = &command_with_login($cmd, $login, $pass);
+my $out = &backquote_logged("$cmd 2>&1");
 if ($? || $out =~ /could not|error|failed/i) {
 	return $out;
 	}
 return undef;
 }
 
+# PostgreSQL versions below 7.3 don't support .pgpass, and version 8.0.*
+# don't allow it to be located via $HOME or $PGPASSFILE.
 sub supports_pgpass
 {
-return &get_postgresql_version(1) >= 7.3;
+local $ver = &get_postgresql_version(1);
+return $ver >= 7.3 && $ver < 8.0 ||
+       $ver >= 8.1;
 }
 
 # command_with_login(command, [user, pass])
@@ -1129,13 +1218,26 @@ if (!defined($user)) {
 local $loginfile;
 if (&supports_pgpass()) {
 	# Can use .pgpass file
-	local $temphome = &transname();
-	&make_dir($temphome, 0755);
-	local $pgpass = "$temphome/.pgpass";
+	local $pgpass;
+	if ($gconfig{'os_type'} eq 'windows') {
+		# On Windows, the file is under ~/application data
+		local $appdata = "$ENV{'HOME'}/application data";
+		&make_dir($appdata, 0755);
+		local $postgresql = "$appdata/postgresql";
+		&make_dir($postgresql, 0755);
+		$pgpass = "$postgresql/pgpass.conf";
+		}
+	else {
+		local $temphome = &transname();
+		&make_dir($temphome, 0755);
+		$pgpass = "$temphome/.pgpass";
+		push(@main::temporary_files, $pgpass);
+		$cmd = "HOME=$temphome $cmd";
+		}
+	$cmd = "PGPASSFILE=$pgpass $cmd";
 	open(PGPASS, ">$pgpass");
 	print PGPASS "*:*:*:$user:$pass\n";
 	close(PGPASS);
-	$ENV{'HOME'} = $temphome;
 	&set_ownership_permissions(
 		$postgres_sameunix ? $user : undef,
 		undef, 0600, $pgpass);
@@ -1151,5 +1253,102 @@ else {
 return $cmd;
 }
 
+# host_port_flags()
+# Returns flags to set the correct host and post for postgreSQL CLI commands
+sub host_port_flags
+{
+local $sslmode = $config{'sslmode'};
+$sslmode =~ s/_/-/g;
+if ($sslmode) {
+	my @rv;
+	push(@rv, "host=".$config{'host'}) if ($config{'host'});
+	push(@rv, "port=".$config{'port'}) if ($config{'port'});
+	push(@rv, "sslmode=".$sslmode) if ($sslmode);
+	return @rv ? " '".join(" ", @rv)."'" : "";
+	}
+else {
+	my $rv = "";
+	$rv .= " -h $config{'host'}" if ($config{'host'});
+	$rv .= " -p $config{'port'}" if ($config{'port'});
+	return $rv;
+	}
+}
+
+# extract_grants(field)
+# Given a field from pg_class that contains grants either as a comma-separated
+# list or an array, return a list of tuples in user,grant format
+sub extract_grants
+{
+my ($f) = @_;
+my @rv;
+if (ref($f)) {
+	@rv = map { [ split(/=/, $_, 2) ] } @$f;
+	}
+else {
+	$f =~ s/^\{//;
+	$f =~ s/\}$//;
+	@rv = map { [ split(/=/, $_, 2) ] } map { s/\\"/"/g; s/"//g; $_ } grep { /=\S/ } split(/,/, $f);
+	}
+return @rv;
+}
+
+# delete_database_backup_job(db)
+# If there is a backup scheduled for some database, remove it
+sub delete_database_backup_job
+{
+my ($db) = @_;
+&foreign_require("cron");
+my @jobs = &cron::list_cron_jobs();
+my $cmd = "$cron_cmd $db";
+my ($job) = grep { $_->{'command'} eq $cmd } @jobs;
+if ($job) {
+	&lock_file(&cron::cron_file($job));
+	&cron::delete_cron_job($job);
+	&unlock_file(&cron::cron_file($job));
+	}
+}
+
+# get_pg_shadow_table()
+# Returns the table containing users, and the list of columns (comma-separated)
+sub get_pg_shadow_table
+{
+if (&get_postgresql_version() >= 9.5) {
+	my $cols = $pg_shadow_cols;
+	$cols =~ s/usecatupd/'t'/g;
+	return ("pg_user", $cols);
+	}
+elsif (&get_postgresql_version() >= 9.4) {
+	return ("pg_user", $pg_shadow_cols);
+	}
+else {
+	return ("pg_shadow", $pg_shadow_cols);
+	}
+}
+
+# compression_format(file)
+# Returns 0 if uncompressed, 1 for gzip, 2 for compress, 3 for bzip2 or
+# 4 for zip
+sub compression_format
+{
+open(BACKUP, "<".$_[0]);
+local $two;
+read(BACKUP, $two, 2);
+close(BACKUP);
+return $two eq "\037\213" ? 1 :
+       $two eq "\037\235" ? 2 :
+       $two eq "PK" ? 4 :
+       $two eq "BZ" ? 3 : 0;
+}
+
+# set_login_pass(same-unix, login, password)
+# Sets the credentials to be used for all SQL commands
+sub set_login_pass
+{
+my ($sameunix, $login, $pass) = @_;
+$postgres_sameunix = $sameunix;
+$postgres_login = $login;
+$postgres_pass = $pass;
+}
+
 1;
 
